@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
@@ -56,14 +57,21 @@ iscsi_scsi_response_cb(struct iscsi_context *iscsi, int status,
 	case SCSI_STATUS_TASK_ABORTED:
 	case SCSI_STATUS_ERROR:
 	case SCSI_STATUS_CANCELLED:
-		scsi_cbdata->callback(iscsi, status, scsi_cbdata->task,
-				      scsi_cbdata->private_data);
+	case SCSI_STATUS_TIMEOUT:
+		scsi_cbdata->task->status = status;
+		if (scsi_cbdata->callback) {
+			scsi_cbdata->callback(iscsi, status, scsi_cbdata->task,
+			                      scsi_cbdata->private_data);
+		}
 		return;
 	default:
+		scsi_cbdata->task->status = SCSI_STATUS_ERROR;
 		iscsi_set_error(iscsi, "Cant handle  scsi status %d yet.",
-				status);
-		scsi_cbdata->callback(iscsi, SCSI_STATUS_ERROR, scsi_cbdata->task,
-				      scsi_cbdata->private_data);
+		                status);
+		if (scsi_cbdata->callback) {
+			scsi_cbdata->callback(iscsi, SCSI_STATUS_ERROR, scsi_cbdata->task,
+			                      scsi_cbdata->private_data);
+		}
 	}
 }
 
@@ -76,24 +84,17 @@ iscsi_send_data_out(struct iscsi_context *iscsi, struct iscsi_pdu *cmd_pdu,
 		struct iscsi_pdu *pdu;
 		int flags;
 
-		if (len > iscsi->target_max_recv_data_segment_length) {
-			len = iscsi->target_max_recv_data_segment_length;
-		}
+		len = MIN(len, iscsi->target_max_recv_data_segment_length);
 
-		pdu = iscsi_allocate_pdu_with_itt_flags(iscsi, ISCSI_PDU_DATA_OUT,
-				 ISCSI_PDU_NO_PDU,
-				 cmd_pdu->itt,
-				 ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_DELETE_WHEN_SENT|ISCSI_PDU_NO_CALLBACK);
+		pdu = iscsi_allocate_pdu(iscsi,
+					 ISCSI_PDU_DATA_OUT,
+					 ISCSI_PDU_NO_PDU,
+					 cmd_pdu->itt,
+					 ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_DELETE_WHEN_SENT);
 		if (pdu == NULL) {
 			iscsi_set_error(iscsi, "Out-of-memory, Failed to allocate "
 				"scsi data out pdu.");
-			ISCSI_LIST_REMOVE(&iscsi->outqueue, cmd_pdu);
-			ISCSI_LIST_REMOVE(&iscsi->waitpdu, cmd_pdu);
-			cmd_pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-				     cmd_pdu->private_data);
-			iscsi_free_pdu(iscsi, cmd_pdu);
-			return -1;
-
+			goto error;
 		}
 		pdu->scsi_cbdata.task         = cmd_pdu->scsi_cbdata.task;
 		/* set the cmdsn in the pdu struct so we can compare with
@@ -116,9 +117,6 @@ iscsi_send_data_out(struct iscsi_context *iscsi, struct iscsi_pdu *cmd_pdu,
 		/* ttt */
 		iscsi_pdu_set_ttt(pdu, ttt);
 
-		/* exp statsn */
-		iscsi_pdu_set_expstatsn(pdu, iscsi->statsn+1);
-
 		/* data sn */
 		iscsi_pdu_set_datasn(pdu, cmd_pdu->datasn++);
 
@@ -131,65 +129,36 @@ iscsi_send_data_out(struct iscsi_context *iscsi, struct iscsi_pdu *cmd_pdu,
 		/* update data segment length */
 		scsi_set_uint32(&pdu->outdata.data[4], pdu->payload_len);
 
-		pdu->callback     = cmd_pdu->callback;
-		pdu->private_data = cmd_pdu->private_data;
-
 		if (iscsi_queue_pdu(iscsi, pdu) != 0) {
 			iscsi_set_error(iscsi, "Out-of-memory: failed to queue iscsi "
 				"scsi pdu.");
-			ISCSI_LIST_REMOVE(&iscsi->outqueue, cmd_pdu);
-			ISCSI_LIST_REMOVE(&iscsi->waitpdu, cmd_pdu);
-			cmd_pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-				     cmd_pdu->private_data);
-			iscsi_free_pdu(iscsi, cmd_pdu);
-			iscsi_free_pdu(iscsi, pdu);
-			return -1;
+			goto error;
 		}
 
 		tot_len -= len;
 		offset  += len;
 	}
 	return 0;
-}
 
-void
-iscsi_timeout_scan(struct iscsi_context *iscsi)
-{
-	struct iscsi_pdu *pdu;
-	struct iscsi_pdu *next_pdu;
-	time_t t = time(NULL);
-
-	for (pdu = iscsi->waitpdu; pdu; pdu = next_pdu) {
-		struct iscsi_scsi_cbdata *scsi_cbdata;
-		struct scsi_task *task;
-
-		next_pdu = pdu->next;
-
-		if (pdu->scsi_timeout == 0) {
-			/* no timeout for this pdu */
-			continue;
-		}
-		if (t < pdu->scsi_timeout) {
-			/* not expired yet */
-			continue;
-		}
-		if (pdu->outdata.data[0] != ISCSI_PDU_SCSI_REQUEST) {
-			continue;
-		}
-
-		scsi_cbdata = &pdu->scsi_cbdata;
-		task = scsi_cbdata->task;
-
-		ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
-		pdu->callback(iscsi, SCSI_STATUS_TIMEOUT,
-				task, pdu->private_data);
-		iscsi_set_error(iscsi, "SCSI command timed out");
-
-		/* task is freed by the sync caller */
-		task->status = SCSI_STATUS_TIMEOUT;
+error:
+	ISCSI_LIST_REMOVE(&iscsi->outqueue, cmd_pdu);
+	ISCSI_LIST_REMOVE(&iscsi->waitpdu, cmd_pdu);
+	if (cmd_pdu->callback) {
+		cmd_pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+						  cmd_pdu->private_data);
 	}
+	iscsi_free_pdu(iscsi, cmd_pdu);
+	return -1;
 }
 
+static int
+iscsi_send_unsolicited_data_out(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
+{
+	uint32_t len = MIN(pdu->expxferlen, iscsi->first_burst_length) - pdu->payload_len;
+
+	return iscsi_send_data_out(iscsi, pdu, 0xffffffff,
+				   pdu->payload_len, len);
+}
 
 /* Using 'struct iscsi_data *d' for data-out is optional
  * and will be converted into a one element data-out iovector.
@@ -202,6 +171,11 @@ iscsi_scsi_command_async(struct iscsi_context *iscsi, int lun,
 	struct iscsi_pdu *pdu;
 	int flags;
 
+	if (iscsi->old_iscsi) {
+		iscsi = iscsi->old_iscsi;
+		ISCSI_LOG(iscsi, 2, "iscsi_scsi_command_async: queuing cmd to old_iscsi while reconnecting");
+	}
+
 	if (iscsi->session_type != ISCSI_SESSION_NORMAL) {
 		iscsi_set_error(iscsi, "Trying to send command on "
 				"discovery session.");
@@ -229,8 +203,11 @@ iscsi_scsi_command_async(struct iscsi_context *iscsi, int lun,
 		scsi_task_set_iov_out(task, iov, 1);
 	}
 
-	pdu = iscsi_allocate_pdu(iscsi, ISCSI_PDU_SCSI_REQUEST,
-				 ISCSI_PDU_SCSI_RESPONSE);
+	pdu = iscsi_allocate_pdu(iscsi,
+				 ISCSI_PDU_SCSI_REQUEST,
+				 ISCSI_PDU_SCSI_RESPONSE,
+				 iscsi_itt_post_increment(iscsi),
+				 0);
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory, Failed to allocate "
 				"scsi pdu.");
@@ -261,13 +238,8 @@ iscsi_scsi_command_async(struct iscsi_context *iscsi, int lun,
 		if (iscsi->use_immediate_data == ISCSI_IMMEDIATE_DATA_YES) {
 			uint32_t len = task->expxferlen;
 
-			if (len > iscsi->first_burst_length) {
-				len = iscsi->first_burst_length;
-			}
-
-			if (len > iscsi->target_max_recv_data_segment_length) {
-				len = iscsi->target_max_recv_data_segment_length;
-			}
+			len = MIN(len, iscsi->first_burst_length);
+			len = MIN(len, iscsi->target_max_recv_data_segment_length);
 
 			pdu->payload_offset = 0;
 			pdu->payload_len    = len;
@@ -300,13 +272,8 @@ iscsi_scsi_command_async(struct iscsi_context *iscsi, int lun,
 	iscsi_pdu_set_expxferlen(pdu, task->expxferlen);
 
 	/* cmdsn */
-	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
-	pdu->cmdsn = iscsi->cmdsn;
-	iscsi->cmdsn++;
+	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn++);
 
-	/* exp statsn */
-	iscsi_pdu_set_expstatsn(pdu, iscsi->statsn+1);
-	
 	/* cdb */
 	iscsi_pdu_set_cdb(pdu, task);
 
@@ -326,13 +293,7 @@ iscsi_scsi_command_async(struct iscsi_context *iscsi, int lun,
 	 * subtract from first_burst_length.
 	 */
 	if (!(flags & ISCSI_PDU_SCSI_FINAL)) {
-		uint32_t len = task->expxferlen;
-
-		if (len + pdu->payload_len > iscsi->first_burst_length) {
-			len = iscsi->first_burst_length - pdu->payload_len;
-		}
-		iscsi_send_data_out(iscsi, pdu, 0xffffffff,
-				    pdu->payload_len, len);
+		iscsi_send_unsolicited_data_out(iscsi, pdu);
 	}
 
 	/* remember cmdsn and itt so we can use task management */
@@ -343,33 +304,114 @@ iscsi_scsi_command_async(struct iscsi_context *iscsi, int lun,
 	return 0;
 }
 
+/* Parse a sense key specific sense data descriptor */
+static void parse_sense_spec(struct scsi_sense *sense, const uint8_t inf[3])
+{
+	if (!(inf[0] & 0x80)) /* SKSV */
+		return;
+
+	sense->sense_specific = 1;
+	sense->ill_param_in_cdb = !!(inf[0] & 0x40); /* C/D flag */
+	if (inf[0] & 8) { /* BPV */
+		sense->bit_pointer_valid = 1;
+		sense->bit_pointer = inf[0] & 7;
+	}
+	sense->field_pointer = scsi_get_uint16(&inf[1]);
+}
+
+/* Parse descriptor format sense data */
+static void parse_sense_descriptors(struct scsi_sense *sense, const uint8_t *sb,
+				    unsigned sb_len)
+{
+	const unsigned char *p, *const end = sb + sb_len;
+
+	for (p = sb; p < end; p += p[1]) {
+		if (p[1] < 4) /* length */
+			break;
+		if (!(p[2] & 0x80)) /* VALID bit */
+			break;
+		switch (p[0]) {
+		case 2:
+			/* Sense key specific sense data descriptor */
+			parse_sense_spec(sense, p + 4);
+			break;
+		}
+	}
+}
+
+void scsi_parse_sense_data(struct scsi_sense *sense, const uint8_t *sb)
+{
+	sense->error_type = sb[0] & 0x7f;
+	switch (sense->error_type) {
+	case 0x70:
+	case 0x71:
+		/* Fixed format */
+		sense->key  = sb[2] & 0x0f;
+		sense->ascq = scsi_get_uint16(&sb[12]);
+		parse_sense_spec(sense, sb + 15);
+		break;
+	case 0x72:
+	case 0x73:
+		/* Descriptor format */
+		sense->key  = sb[1] & 0x0f;
+		sense->ascq = scsi_get_uint16(&sb[2]);
+		parse_sense_descriptors(sense, sb + 8, sb[7]);
+		break;
+	}
+}
+
 int
 iscsi_process_scsi_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 			 struct iscsi_in_pdu *in)
 {
-	uint32_t flags, status;
+	uint8_t flags, response, status;
 	struct iscsi_scsi_cbdata *scsi_cbdata = &pdu->scsi_cbdata;
 	struct scsi_task *task = scsi_cbdata->task;
 
-	iscsi_adjust_statsn(iscsi, in);
-	iscsi_adjust_maxexpcmdsn(iscsi, in);
-
 	flags = in->hdr[1];
 	if ((flags&ISCSI_PDU_DATA_FINAL) == 0) {
 		iscsi_set_error(iscsi, "scsi response pdu but Final bit is "
 				"not set: 0x%02x.", flags);
-		pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
+			              pdu->private_data);
+		}
 		return -1;
 	}
 	if ((flags&ISCSI_PDU_DATA_ACK_REQUESTED) != 0) {
 		iscsi_set_error(iscsi, "scsi response asked for ACK "
 				"0x%02x.", flags);
-		pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
+			              pdu->private_data);
+		}
 		return -1;
 	}
 
+	response = in->hdr[2];
+
+	task->residual_status = SCSI_RESIDUAL_NO_RESIDUAL;
+	task->residual = 0;
+
+	if (flags & (ISCSI_PDU_DATA_RESIDUAL_OVERFLOW|
+		     ISCSI_PDU_DATA_RESIDUAL_UNDERFLOW)) {
+		if (response != 0) {
+			iscsi_set_error(iscsi, "protocol error: flags %#02x;"
+					" response %#02x.", flags, response);
+			if (pdu->callback) {
+				pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
+				              pdu->private_data);
+			}
+			return -1;
+		}
+		task->residual = scsi_get_uint32(&in->hdr[44]);
+		if (flags & ISCSI_PDU_DATA_RESIDUAL_UNDERFLOW) {
+			task->residual_status = SCSI_RESIDUAL_UNDERFLOW;
+		} else {
+			task->residual_status = SCSI_RESIDUAL_OVERFLOW;
+		}
+	}
+
 	status = in->hdr[3];
 
 	switch (status) {
@@ -378,21 +420,6 @@ iscsi_process_scsi_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		task->datain.data = pdu->indata.data;
 		task->datain.size = pdu->indata.size;
 
-		task->residual_status = SCSI_RESIDUAL_NO_RESIDUAL;
-		task->residual = 0;
-
-		/*
-		 * These flags should only be set if the S flag is also set
-		 */
-		if (flags & (ISCSI_PDU_DATA_RESIDUAL_OVERFLOW|ISCSI_PDU_DATA_RESIDUAL_UNDERFLOW)) {
-			task->residual = scsi_get_uint32(&in->hdr[44]);
-			if (flags & ISCSI_PDU_DATA_RESIDUAL_UNDERFLOW) {
-				task->residual_status = SCSI_RESIDUAL_UNDERFLOW;
-			} else {
-				task->residual_status = SCSI_RESIDUAL_OVERFLOW;
-			}
-		}
-
 		/* the pdu->datain.data was malloc'ed by iscsi_malloc,
 		   as long as we have no struct iscsi_task we cannot track
 		   the free'ing of this buffer which is currently
@@ -402,8 +429,10 @@ iscsi_process_scsi_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		pdu->indata.data = NULL;
 		pdu->indata.size = 0;
 
-		pdu->callback(iscsi, SCSI_STATUS_GOOD, task,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_GOOD, task,
+			              pdu->private_data);
+		}
 		break;
 	case SCSI_STATUS_CHECK_CONDITION:
 		task->datain.size = in->data_pos;
@@ -415,59 +444,58 @@ iscsi_process_scsi_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		}
 		memcpy(task->datain.data, in->data, task->datain.size);
 
-		task->sense.error_type = task->datain.data[2] & 0x7f;
-		switch (task->sense.error_type) {
-		case 0x70:
-		case 0x71:
-		  task->sense.key        = task->datain.data[4] & 0x0f;
-		  task->sense.ascq       = scsi_get_uint16(
-						&(task->datain.data[14]));
-		  break;
-		case 0x72:
-		case 0x73:
-		  task->sense.key        = task->datain.data[3] & 0x0f;
-		  task->sense.ascq       = scsi_get_uint16(
-						&(task->datain.data[4]));
-		  break;
-		}
+		scsi_parse_sense_data(&task->sense, &task->datain.data[2]);
 		iscsi_set_error(iscsi, "SENSE KEY:%s(%d) ASCQ:%s(0x%04x)",
 				scsi_sense_key_str(task->sense.key),
 				task->sense.key,
 				scsi_sense_ascq_str(task->sense.ascq),
 				task->sense.ascq);
-		pdu->callback(iscsi, SCSI_STATUS_CHECK_CONDITION, task,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_CHECK_CONDITION, task,
+			              pdu->private_data);
+		}
 		break;
 	case SCSI_STATUS_RESERVATION_CONFLICT:
 		iscsi_set_error(iscsi, "RESERVATION CONFLICT");
-		pdu->callback(iscsi, SCSI_STATUS_RESERVATION_CONFLICT,
-			task, pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_RESERVATION_CONFLICT,
+			              task, pdu->private_data);
+		}
 		break;
 	case SCSI_STATUS_TASK_SET_FULL:
 		iscsi_set_error(iscsi, "TASK_SET_FULL");
-		pdu->callback(iscsi, SCSI_STATUS_TASK_SET_FULL,
-			task, pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_TASK_SET_FULL,
+			              task, pdu->private_data);
+		}
 		break;
 	case SCSI_STATUS_ACA_ACTIVE:
 		iscsi_set_error(iscsi, "ACA_ACTIVE");
-		pdu->callback(iscsi, SCSI_STATUS_ACA_ACTIVE,
-			task, pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ACA_ACTIVE,
+			              task, pdu->private_data);
+		}
 		break;
 	case SCSI_STATUS_TASK_ABORTED:
 		iscsi_set_error(iscsi, "TASK_ABORTED");
-		pdu->callback(iscsi, SCSI_STATUS_TASK_ABORTED,
-			task, pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_TASK_ABORTED,
+			              task, pdu->private_data);
+		}
 		break;
 	case SCSI_STATUS_BUSY:
 		iscsi_set_error(iscsi, "BUSY");
-		pdu->callback(iscsi, SCSI_STATUS_BUSY,
-			task, pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_BUSY,
+			              task, pdu->private_data);
+		}
 		break;
 	default:
 		iscsi_set_error(iscsi, "Unknown SCSI status :%d.", status);
-
-		pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR,
+			              task, pdu->private_data);
+		}
 		return -1;
 	}
 
@@ -483,15 +511,14 @@ iscsi_process_scsi_data_in(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	struct scsi_task *task = scsi_cbdata->task;
 	int dsl;
 
-	iscsi_adjust_statsn(iscsi, in);
-	iscsi_adjust_maxexpcmdsn(iscsi, in);
-
 	flags = in->hdr[1];
 	if ((flags&ISCSI_PDU_DATA_ACK_REQUESTED) != 0) {
 		iscsi_set_error(iscsi, "scsi response asked for ACK "
 				"0x%02x.", flags);
-		pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR, task,
+			              pdu->private_data);
+		}
 		return -1;
 	}
 	dsl = scsi_get_uint32(&in->hdr[4]) & 0x00ffffff;
@@ -548,7 +575,9 @@ iscsi_process_scsi_data_in(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	pdu->indata.data = NULL;
 	pdu->indata.size = 0;
 
-	pdu->callback(iscsi, status, task, pdu->private_data);
+	if (pdu->callback) {
+		pdu->callback(iscsi, status, task, pdu->private_data);
+	}
 
 	return 0;
 }
@@ -563,8 +592,6 @@ iscsi_process_r2t(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	offset = scsi_get_uint32(&in->hdr[40]);
 	len    = scsi_get_uint32(&in->hdr[44]);
 
-	iscsi_adjust_maxexpcmdsn(iscsi, in);
-
 	pdu->datasn = 0;
 	iscsi_send_data_out(iscsi, pdu, ttt, offset, len);
 	return 0;
@@ -930,6 +957,40 @@ iscsi_write16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 	return task;
 }
 
+struct scsi_task *
+iscsi_writeatomic16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			 unsigned char *data, uint32_t datalen, int blocksize,
+			 int wrprotect, int dpo, int fua, int group_number,
+			 iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data d;
+
+	if (datalen % blocksize != 0) {
+		iscsi_set_error(iscsi, "Datalen:%d is not a multiple of the "
+				"blocksize:%d.", datalen, blocksize);
+		return NULL;
+	}
+
+	task = scsi_cdb_writeatomic16(lba, datalen, blocksize, wrprotect,
+				      dpo, fua, group_number);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"writeAtomic16 cdb.");
+		return NULL;
+	}
+	d.data = data;
+	d.size = datalen;
+
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb,
+				     &d, private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
 struct scsi_task *
 iscsi_orwrite_task(struct iscsi_context *iscsi, int lun, uint64_t lba, 
 		   unsigned char *data, uint32_t datalen, int blocksize,
@@ -1511,7 +1572,8 @@ iscsi_writesame10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 	struct scsi_task *task;
 	struct iscsi_data d;
 
-	task = scsi_cdb_writesame10(wrprotect, anchor, unmap, lba, group, num_blocks);
+	task = scsi_cdb_writesame10(wrprotect, anchor, unmap, lba, group,
+				    num_blocks, datalen);
 	if (task == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
 				"writesame10 cdb.");
@@ -1544,7 +1606,8 @@ iscsi_writesame16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 	struct scsi_task *task;
 	struct iscsi_data d;
 
-	task = scsi_cdb_writesame16(wrprotect, anchor, unmap, lba, group, num_blocks);
+	task = scsi_cdb_writesame16(wrprotect, anchor, unmap, lba, group,
+				    num_blocks, datalen);
 	if (task == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
 				"writesame16 cdb.");
@@ -1855,7 +1918,7 @@ iscsi_scsi_cancel_task(struct iscsi_context *iscsi,
 	for (pdu = iscsi->waitpdu; pdu; pdu = pdu->next) {
 		if (pdu->itt == task->itt) {
 			ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
-			if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+			if (pdu->callback) {
 				pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
 				      pdu->private_data);
 			}
@@ -1866,7 +1929,7 @@ iscsi_scsi_cancel_task(struct iscsi_context *iscsi,
 	for (pdu = iscsi->outqueue; pdu; pdu = pdu->next) {
 		if (pdu->itt == task->itt) {
 			ISCSI_LIST_REMOVE(&iscsi->outqueue, pdu);
-			if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+			if (pdu->callback) {
 				pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
 				      pdu->private_data);
 			}
@@ -1884,7 +1947,7 @@ iscsi_scsi_cancel_all_tasks(struct iscsi_context *iscsi)
 
 	while ((pdu = iscsi->waitpdu)) {
 		ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
-		if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+		if (pdu->callback) {
 			pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
 				      pdu->private_data);
 		}
@@ -1892,11 +1955,10 @@ iscsi_scsi_cancel_all_tasks(struct iscsi_context *iscsi)
 	}
 	while ((pdu = iscsi->outqueue)) {
 		ISCSI_LIST_REMOVE(&iscsi->outqueue, pdu);
-		if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+		if (pdu->callback) {
 			pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
 				      pdu->private_data);
 		}
 		iscsi_free_pdu(iscsi, pdu);
 	}
 }
-
