@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
@@ -54,8 +55,9 @@ iscsi_login_add_initiatorname(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 	char str[MAX_STRING_SIZE+1];
 
 	/* We only send InitiatorName during opneg or the first leg of secneg */
-	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
-	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
+	if ((iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
+	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP)
+	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
 		return 0;
 	}
 
@@ -77,8 +79,9 @@ iscsi_login_add_alias(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 	char str[MAX_STRING_SIZE+1];
 
 	/* We only send InitiatorAlias during opneg or the first leg of secneg */
-	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
-	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
+	if ((iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
+	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP)
+	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
 		return 0;
 	}
 
@@ -101,8 +104,9 @@ iscsi_login_add_targetname(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 	char str[MAX_STRING_SIZE+1];
 
 	/* We only send TargetName during opneg or the first leg of secneg */
-	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
-	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
+	if ((iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
+	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP)
+	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
 		return 0;
 	}
 
@@ -130,9 +134,10 @@ iscsi_login_add_sessiontype(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
 	char str[MAX_STRING_SIZE+1];
 
-	/* We only send TargetName during opneg or the first leg of secneg */
-	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
-	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
+	/* We only send SessionType during opneg or the first leg of secneg */
+	if ((iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
+	&& iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP)
+	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
 		return 0;
 	}
 
@@ -637,13 +642,16 @@ static void gcry_md_close(gcry_md_hd_t h)
 }
 #endif
 
+/* size of the challenge used for bidirectional chap */
+#define TARGET_CHAP_C_SIZE 32
+
 static int
 iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
 	char str[MAX_STRING_SIZE+1];
 	char * strp;
 	unsigned char c, cc[2];
-	unsigned char digest[16];
+	unsigned char digest[CHAP_R_SIZE];
 	gcry_md_hd_t ctx;
 	int i;
 
@@ -652,14 +660,14 @@ iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 		return 0;
 	}
 
-	gcry_md_open(&ctx, GCRY_MD_MD5, 0);
-	if (!ctx) {
-		iscsi_set_error(iscsi, "Cannot create MD5 algorithm");
+	if (!iscsi->chap_c[0]) {
+		iscsi_set_error(iscsi, "No CHAP challenge found");
 		return -1;
 	}
 
-	if (!iscsi->chap_c[0]) {
-		iscsi_set_error(iscsi, "No CHAP challenge found");
+	gcry_md_open(&ctx, GCRY_MD_MD5, 0);
+	if (ctx == NULL) {
+		iscsi_set_error(iscsi, "Cannot create MD5 algorithm");
 		return -1;
 	}
 	gcry_md_putc(ctx, iscsi->chap_i);
@@ -681,7 +689,7 @@ iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 		return -1;
 	}
 
-	for (i=0; i<16; i++) {
+	for (i = 0; i < CHAP_R_SIZE; i++) {
 		c = digest[i];
 		cc[0] = i2h((c >> 4)&0x0f);
 		cc[1] = i2h((c     )&0x0f);
@@ -691,6 +699,7 @@ iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 			return -1;
 		}
 	}
+
 	c = 0;
 	if (iscsi_pdu_add_data(iscsi, pdu, &c, 1) != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data "
@@ -698,6 +707,63 @@ iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 		return -1;
 	}
 
+	/* bidirectional chap */
+	if (iscsi->target_user[0]) {
+		unsigned char target_chap_c[TARGET_CHAP_C_SIZE];
+
+		iscsi->target_chap_i++;
+		snprintf(str, MAX_STRING_SIZE, "CHAP_I=%d",
+			 iscsi->target_chap_i);
+		if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str,
+				       strlen(str) + 1) != 0) {
+			iscsi_set_error(iscsi, "Out-of-memory: pdu add "
+					"data failed.");
+			return -1;
+		}
+
+		for (i = 0; i < TARGET_CHAP_C_SIZE; i++) {
+			target_chap_c[i] = rand()&0xff;
+		}
+		strncpy(str, "CHAP_C=0x", MAX_STRING_SIZE);
+		if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str,
+				       strlen(str)) != 0) {
+			iscsi_set_error(iscsi, "Out-of-memory: pdu add data "
+					"failed.");
+			return -1;
+		}
+		for (i = 0; i < TARGET_CHAP_C_SIZE; i++) {
+			c = target_chap_c[i];
+			cc[0] = i2h((c >> 4)&0x0f);
+			cc[1] = i2h((c     )&0x0f);
+			if (iscsi_pdu_add_data(iscsi, pdu, &cc[0], 2) != 0) {
+				iscsi_set_error(iscsi, "Out-of-memory: pdu add "
+						"data failed.");
+				return -1;
+			}
+		}
+		c = 0;
+		if (iscsi_pdu_add_data(iscsi, pdu, &c, 1) != 0) {
+			iscsi_set_error(iscsi, "Out-of-memory: pdu add data "
+					"failed.");
+			return -1;
+		}
+
+		gcry_md_open(&ctx, GCRY_MD_MD5, 0);
+		if (ctx == NULL) {
+			iscsi_set_error(iscsi, "Cannot create MD5 algorithm");
+			return -1;
+		}
+		gcry_md_putc(ctx, iscsi->target_chap_i);
+		gcry_md_write(ctx, (unsigned char *)iscsi->target_passwd,
+			      strlen(iscsi->target_passwd));
+		gcry_md_write(ctx, (unsigned char *)target_chap_c,
+			      TARGET_CHAP_C_SIZE);
+
+		memcpy(iscsi->target_chap_r, gcry_md_read(ctx, 0),
+		       sizeof(iscsi->target_chap_r));
+		gcry_md_close(ctx);
+	}
+
 	return 0;
 }
 
@@ -730,10 +796,18 @@ iscsi_login_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 		return -1;
 	}
 
-	pdu = iscsi_allocate_pdu_with_itt_flags(iscsi,
-				ISCSI_PDU_LOGIN_REQUEST,
-				ISCSI_PDU_LOGIN_RESPONSE,
-				iscsi->itt, 0);
+	/* randomize cmdsn and itt */
+	if (!iscsi->current_phase && !iscsi->secneg_phase) {
+		iscsi->itt = (uint32_t) rand();
+		iscsi->cmdsn = (uint32_t) rand();
+		iscsi->expcmdsn = iscsi->maxcmdsn = iscsi->min_cmdsn_waiting = iscsi->cmdsn;
+	}
+
+	pdu = iscsi_allocate_pdu(iscsi,
+				 ISCSI_PDU_LOGIN_REQUEST,
+				 ISCSI_PDU_LOGIN_RESPONSE,
+				 iscsi->itt,
+				 ISCSI_PDU_DROP_ON_RECONNECT);
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory: Failed to allocate "
 				"login pdu.");
@@ -743,6 +817,9 @@ iscsi_login_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	/* login request */
 	iscsi_pdu_set_immediate(pdu);
 
+	/* cmdsn is not increased if Immediate delivery*/
+	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
+
 	if (!iscsi->user[0]) {
 		iscsi->current_phase = ISCSI_PDU_LOGIN_CSG_OPNEG;
 	}
@@ -968,11 +1045,27 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	uint32_t status;
 	char *ptr = (char *)in->data;
 	int size = in->data_pos;
+	int must_have_chap_n = 0;
+	int must_have_chap_r = 0;
 
 	status = scsi_get_uint16(&in->hdr[36]);
 
-	iscsi_adjust_statsn(iscsi, in);
-	iscsi_adjust_maxexpcmdsn(iscsi, in);
+	// Status-Class is 0
+	if (!(status >> 8)) {
+		if (!iscsi->current_phase && !iscsi->secneg_phase) {
+			iscsi->statsn = scsi_get_uint32(&in->hdr[24]);
+		}
+	}
+
+	/* Using bidirectional CHAP? Then we must see a chap_n and chap_r
+	 * field in this PDU
+	 */
+	if ((in->hdr[1] & ISCSI_PDU_LOGIN_TRANSIT)
+	&& (in->hdr[1] & ISCSI_PDU_LOGIN_CSG_FF) == ISCSI_PDU_LOGIN_CSG_SECNEG
+	&& iscsi->target_user[0]) {
+		must_have_chap_n = 1;
+		must_have_chap_r = 1;
+	}
 
 	/* XXX here we should parse the data returned in case the target
 	 * renegotiated some some parameters.
@@ -989,8 +1082,10 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 			iscsi_set_error(iscsi, "NUL not found after offset %ld "
 					"when parsing login data",
 					(long)((unsigned char *)ptr - in->data));
-			pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-				      pdu->private_data);
+			if (pdu->callback) {
+				pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+				              pdu->private_data);
+			}
 			return -1;
 		}
 
@@ -1065,26 +1160,103 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		}
 
 		if (!strncmp(ptr, "CHAP_C=0x", 9)) {
-			strncpy(iscsi->chap_c,ptr+9,MAX_STRING_SIZE);
+			if (len-9 > MAX_CHAP_C_LENGTH) {
+				iscsi_set_error(iscsi, "Wrong length of CHAP_C received from"
+						" target (%d, max: %d)", len-9, MAX_CHAP_C_LENGTH);
+				if (pdu->callback) {
+					pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+					              pdu->private_data);
+				}
+				return 0;
+			}
+			*iscsi->chap_c = '\0';
+			strncat(iscsi->chap_c,ptr+9,len-9);
 			iscsi->secneg_phase = ISCSI_LOGIN_SECNEG_PHASE_SEND_RESPONSE;
 		}
 
+		if (!strncmp(ptr, "CHAP_N=", 7)) {
+			if (strcmp(iscsi->target_user, ptr + 7)) {
+				iscsi_set_error(iscsi, "Failed to log in to"
+						" target. Wrong CHAP targetname"
+						" received: %s", ptr + 7);
+				if (pdu->callback) {
+					pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+					              pdu->private_data);
+				}
+				return 0;
+			}
+			must_have_chap_n = 0;
+		}
+
+		if (!strncmp(ptr, "CHAP_R=0x", 9)) {
+			int i;
+
+			if (len != 9 + 2 * CHAP_R_SIZE) {
+				iscsi_set_error(iscsi, "Wrong size of CHAP_R"
+						" received from target.");
+				if (pdu->callback) {
+					pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+					              pdu->private_data);
+				}
+				return 0;
+			}
+			for (i = 0; i < CHAP_R_SIZE; i++) {
+				unsigned char c;
+				c = ((h2i(ptr[9 + 2 * i]) << 4) | h2i(ptr[9 + 2 * i + 1]));
+				if (c != iscsi->target_chap_r[i]) {
+					iscsi_set_error(iscsi, "Authentication "
+						"failed. Invalid CHAP_R "
+						"response from the target");
+					if (pdu->callback) {
+						pdu->callback(iscsi, SCSI_STATUS_ERROR,
+						              NULL, pdu->private_data);
+					}
+					return 0;
+				}
+			}
+			must_have_chap_r = 0;
+		}
+
 		ptr  += len + 1;
 		size -= len + 1;
 	}
 
 	if (status == SCSI_STATUS_REDIRECT && iscsi->target_address[0]) {
 		ISCSI_LOG(iscsi, 2, "target requests redirect to %s",iscsi->target_address);
-		pdu->callback(iscsi, SCSI_STATUS_REDIRECT, NULL,
-				  pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_REDIRECT, NULL,
+			              pdu->private_data);
+		}
 		return 0;
 	}
 
 	if (status != 0) {
 		iscsi_set_error(iscsi, "Failed to log in to target. Status: %s(%d)",
 				       login_error_str(status), status);
-		pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-			      pdu->private_data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+			              pdu->private_data);
+		}
+		return 0;
+	}
+
+	if (must_have_chap_n) {
+		iscsi_set_error(iscsi, "Failed to log in to target. "
+				"It did not return CHAP_N during SECNEG.");
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+			              pdu->private_data);
+		}
+		return 0;
+	}
+
+	if (must_have_chap_r) {
+		iscsi_set_error(iscsi, "Failed to log in to target. "
+				"It did not return CHAP_R during SECNEG.");
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+			              pdu->private_data);
+		}
 		return 0;
 	}
 
@@ -1102,7 +1274,9 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	} else {
 		if (iscsi_login_async(iscsi, pdu->callback, pdu->private_data) != 0) {
 			iscsi_set_error(iscsi, "Failed to send continuation login pdu");
-			pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL, pdu->private_data);
+			if (pdu->callback) {
+				pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL, pdu->private_data);
+			}
 		}
 	}
 
@@ -1110,8 +1284,8 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 }
 
 int
-iscsi_logout_async_internal(struct iscsi_context *iscsi, iscsi_command_cb cb,
-		   void *private_data, uint32_t flags)
+iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
+		   void *private_data)
 {
 	struct iscsi_pdu *pdu;
 
@@ -1122,8 +1296,11 @@ iscsi_logout_async_internal(struct iscsi_context *iscsi, iscsi_command_cb cb,
 		return -1;
 	}
 
-	pdu = iscsi_allocate_pdu(iscsi, ISCSI_PDU_LOGOUT_REQUEST,
-				 ISCSI_PDU_LOGOUT_RESPONSE);
+	pdu = iscsi_allocate_pdu(iscsi,
+				 ISCSI_PDU_LOGOUT_REQUEST,
+				 ISCSI_PDU_LOGOUT_RESPONSE,
+				 iscsi_itt_post_increment(iscsi),
+				 ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_CORK_WHEN_SENT);
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory: Failed to allocate "
 				"logout pdu.");
@@ -1138,14 +1315,9 @@ iscsi_logout_async_internal(struct iscsi_context *iscsi, iscsi_command_cb cb,
 
 	/* cmdsn is not increased if Immediate delivery*/
 	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
-	pdu->cmdsn = iscsi->cmdsn;
-
-	/* exp statsn */
-	iscsi_pdu_set_expstatsn(pdu, iscsi->statsn+1);
 
 	pdu->callback     = cb;
 	pdu->private_data = private_data;
-	pdu->flags |= ISCSI_PDU_CORK_WHEN_SENT | flags;
 
 	if (iscsi_queue_pdu(iscsi, pdu) != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: failed to queue iscsi "
@@ -1157,21 +1329,15 @@ iscsi_logout_async_internal(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	return 0;
 }
 
-int
-iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
-		   void *private_data)
-{
-	return iscsi_logout_async_internal(iscsi, cb, private_data, 0);
-}
-
 int
 iscsi_process_logout_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
-struct iscsi_in_pdu *in)
+struct iscsi_in_pdu *in _U_)
 {
-	iscsi_adjust_maxexpcmdsn(iscsi, in);
-
 	iscsi->is_loggedin = 0;
-	pdu->callback(iscsi, SCSI_STATUS_GOOD, NULL, pdu->private_data);
+	ISCSI_LOG(iscsi, 2, "logout successful");
+	if (pdu->callback) {
+		pdu->callback(iscsi, SCSI_STATUS_GOOD, NULL, pdu->private_data);
+	}
 
 	return 0;
 }
