@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
@@ -21,6 +22,7 @@
 #endif
 
 #include <stdio.h>
+#include <stdlib.h>
 #include "iscsi.h"
 #include "iscsi-private.h"
 
@@ -30,24 +32,29 @@ iscsi_nop_out_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 {
 	struct iscsi_pdu *pdu;
 
+	if (iscsi->old_iscsi || iscsi->pending_reconnect) {
+		ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
+		    "NOP Out Send NOT SEND while reconnecting (nops_in_flight: %d, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x)",
+		    iscsi->nops_in_flight, iscsi->maxcmdsn, iscsi->expcmdsn);
+		return 0;
+	}
+
 	if (iscsi->is_loggedin == 0) {
 		iscsi_set_error(iscsi, "trying to send nop-out while not "
 				"logged in");
 		return -1;
 	}
 
-	pdu = iscsi_allocate_pdu(iscsi, ISCSI_PDU_NOP_OUT, ISCSI_PDU_NOP_IN);
+	pdu = iscsi_allocate_pdu(iscsi,
+				 ISCSI_PDU_NOP_OUT,
+				 ISCSI_PDU_NOP_IN,
+				 iscsi_itt_post_increment(iscsi),
+				 ISCSI_PDU_DROP_ON_RECONNECT);
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "Failed to allocate nop-out pdu");
 		return -1;
 	}
 
-	/* We don't want to requeue these on reconnect */
-	pdu->flags |= ISCSI_PDU_DROP_ON_RECONNECT;
-
-	/* immediate flag */
-	iscsi_pdu_set_immediate(pdu);
-
 	/* flags */
 	iscsi_pdu_set_pduflags(pdu, 0x80);
 
@@ -57,12 +64,8 @@ iscsi_nop_out_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	/* lun */
 	iscsi_pdu_set_lun(pdu, 0);
 
-	/* cmdsn is not increased if Immediate delivery*/
-	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
-	pdu->cmdsn = iscsi->cmdsn;
-
-	/* exp statsn */
-	iscsi_pdu_set_expstatsn(pdu, iscsi->statsn+1);
+	/* cmdsn */
+	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn++);
 
 	pdu->callback     = cb;
 	pdu->private_data = private_data;
@@ -82,17 +85,23 @@ iscsi_nop_out_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	}
 
 	iscsi->nops_in_flight++;
+	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
+	          "NOP Out Send (nops_in_flight: %d, pdu->cmdsn %08x, pdu->itt %08x, pdu->ttt %08x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x)",
+	          iscsi->nops_in_flight, pdu->cmdsn, pdu->itt, 0xffffffff, iscsi->maxcmdsn, iscsi->expcmdsn);
 
 	return 0;
 }
 
 int
-iscsi_send_target_nop_out(struct iscsi_context *iscsi, uint32_t ttt)
+iscsi_send_target_nop_out(struct iscsi_context *iscsi, uint32_t ttt, uint32_t lun)
 {
 	struct iscsi_pdu *pdu;
 
-	pdu = iscsi_allocate_pdu_with_itt_flags(iscsi, ISCSI_PDU_NOP_OUT, ISCSI_PDU_NO_PDU,
-				0xffffffff,ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_DELETE_WHEN_SENT|ISCSI_PDU_NO_CALLBACK);
+	pdu = iscsi_allocate_pdu(iscsi,
+				 ISCSI_PDU_NOP_OUT,
+				 ISCSI_PDU_NO_PDU,
+				 0xffffffff,
+				 ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_DELETE_WHEN_SENT);
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "Failed to allocate nop-out pdu");
 		return -1;
@@ -108,14 +117,10 @@ iscsi_send_target_nop_out(struct iscsi_context *iscsi, uint32_t ttt)
 	iscsi_pdu_set_ttt(pdu, ttt);
 
 	/* lun */
-	iscsi_pdu_set_lun(pdu, 0);
+	iscsi_pdu_set_lun(pdu, lun);
 
 	/* cmdsn is not increased if Immediate delivery*/
 	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
-	pdu->cmdsn = iscsi->cmdsn;
-
-	/* exp statsn */
-	iscsi_pdu_set_expstatsn(pdu, iscsi->statsn+1);
 
 	if (iscsi_queue_pdu(iscsi, pdu) != 0) {
 		iscsi_set_error(iscsi, "failed to queue iscsi nop-out pdu");
@@ -123,6 +128,10 @@ iscsi_send_target_nop_out(struct iscsi_context *iscsi, uint32_t ttt)
 		return -1;
 	}
 
+	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
+	          "NOP Out Send (nops_in_flight: %d, pdu->cmdsn %08x, pdu->itt %08x, pdu->ttt %08x, pdu->lun %8x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x)",
+	          iscsi->nops_in_flight, pdu->cmdsn, 0xffffffff, ttt, lun, iscsi->maxcmdsn, iscsi->expcmdsn);
+
 	return 0;
 }
 
@@ -132,7 +141,20 @@ iscsi_process_nop_out_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 {
 	struct iscsi_data data;
 
-	iscsi->nops_in_flight = 0;
+	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
+	          "NOP-In received (pdu->itt %08x, pdu->ttt %08x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x, iscsi->statsn %08x)",
+	          pdu->itt, 0xffffffff, iscsi->maxcmdsn, iscsi->expcmdsn, iscsi->statsn); 
+
+	if (iscsi->waitpdu->cmdsn == iscsi->min_cmdsn_waiting) {
+		ISCSI_LOG(iscsi, 2, "Oldest element in waitqueue is unchanged since last NOP-In (iscsi->min_cmdsn_waiting %08x)",
+		          iscsi->min_cmdsn_waiting); 
+		if (getenv("LIBISCSI_IGNORE_NOP_OUT_ON_STUCK_WAITPDU_QUEUE") == NULL) {
+			iscsi->nops_in_flight = 0;
+		}
+	} else {
+		iscsi->nops_in_flight = 0;
+	}
+	iscsi->min_cmdsn_waiting = iscsi->waitpdu->cmdsn;
 
 	if (pdu->callback == NULL) {
 		return 0;
