@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
@@ -69,43 +70,19 @@ iscsi_itt_post_increment(struct iscsi_context *iscsi) {
 	return old_itt;
 }
 
-void iscsi_adjust_statsn(struct iscsi_context *iscsi, struct iscsi_in_pdu *in) {
-	uint32_t statsn = scsi_get_uint32(&in->hdr[24]);
-	uint32_t itt = scsi_get_uint32(&in->hdr[16]);
-	
-	if (itt == 0xffffffff) {
-		/* target will not increase statsn if itt == 0xffffffff */
-		statsn--;
-	}
-	
-	if (iscsi_serial32_compare(statsn, iscsi->statsn) > 0) {
-		iscsi->statsn = statsn;
-	}
-}
-
-void iscsi_adjust_maxexpcmdsn(struct iscsi_context *iscsi, struct iscsi_in_pdu *in) {
-	uint32_t maxcmdsn = scsi_get_uint32(&in->hdr[32]);
-	if (iscsi_serial32_compare(maxcmdsn, iscsi->maxcmdsn) > 0) {
-		iscsi->maxcmdsn = maxcmdsn;
-	}
-	uint32_t expcmdsn = scsi_get_uint32(&in->hdr[28]);
-	if (iscsi_serial32_compare(expcmdsn, iscsi->expcmdsn) > 0) {
-		iscsi->expcmdsn = expcmdsn;
-	}
-}
-
 void iscsi_dump_pdu_header(struct iscsi_context *iscsi, unsigned char *data) {
 	char dump[ISCSI_RAW_HEADER_SIZE*3+1]={0};
 	int i;
 	for (i=0;i<ISCSI_RAW_HEADER_SIZE;i++) {
 		snprintf(&dump[i * 3], 4, " %02x", data[i]);
 	}
-	ISCSI_LOG(iscsi, 0, "PDU header:%s",dump);
+	ISCSI_LOG(iscsi, 2, "PDU header:%s", dump);
 }
 
 struct iscsi_pdu *
-iscsi_allocate_pdu_with_itt_flags(struct iscsi_context *iscsi, enum iscsi_opcode opcode,
-				  enum iscsi_opcode response_opcode, uint32_t itt, uint32_t flags)
+iscsi_allocate_pdu(struct iscsi_context *iscsi, enum iscsi_opcode opcode,
+		   enum iscsi_opcode response_opcode, uint32_t itt,
+		   uint32_t flags)
 {
 	struct iscsi_pdu *pdu;
 
@@ -143,13 +120,6 @@ iscsi_allocate_pdu_with_itt_flags(struct iscsi_context *iscsi, enum iscsi_opcode
 	return pdu;
 }
 
-struct iscsi_pdu *
-iscsi_allocate_pdu(struct iscsi_context *iscsi, enum iscsi_opcode opcode,
-		   enum iscsi_opcode response_opcode)
-{
-	return iscsi_allocate_pdu_with_itt_flags(iscsi, opcode, response_opcode, iscsi_itt_post_increment(iscsi), 0);
-}	
-
 void
 iscsi_free_pdu(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
@@ -325,18 +295,20 @@ static const char *iscsi_reject_reason_str(enum iscsi_reject_reason reason)
 int iscsi_process_target_nop_in(struct iscsi_context *iscsi,
 				struct iscsi_in_pdu *in)
 {
-	uint32_t ttt;
+	uint32_t ttt = scsi_get_uint32(&in->hdr[20]);
+	uint32_t itt = scsi_get_uint32(&in->hdr[16]);
+	uint32_t lun = scsi_get_uint16(&in->hdr[8]);
 
-	ttt = scsi_get_uint32(&in->hdr[20]);
-
-	iscsi_adjust_statsn(iscsi, in);
+	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
+	          "NOP-In received (pdu->itt %08x, pdu->ttt %08x, pdu->lun %8x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x, iscsi->statsn %08x)",
+	          itt, ttt, lun, iscsi->maxcmdsn, iscsi->expcmdsn, iscsi->statsn);
 
 	/* if the server does not want a response */
 	if (ttt == 0xffffffff) {
 		return 0;
 	}
 
-	iscsi_send_target_nop_out(iscsi, ttt);
+	iscsi_send_target_nop_out(iscsi, ttt, lun);
 
 	return 0;
 }
@@ -347,13 +319,10 @@ static void iscsi_reconnect_after_logout(struct iscsi_context *iscsi, int status
 {
 	if (status) {
 		ISCSI_LOG(iscsi, 1, "logout failed: %s", iscsi_get_error(iscsi));
-	} else {
-		ISCSI_LOG(iscsi, 2, "logout was successful");
 	}
 	iscsi->pending_reconnect = 1;
 }
 
-
 int iscsi_process_reject(struct iscsi_context *iscsi,
 				struct iscsi_in_pdu *in)
 {
@@ -371,7 +340,7 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 
 	if (reason == ISCSI_REJECT_WAITING_FOR_LOGOUT) {
 		ISCSI_LOG(iscsi, 1, "target rejects request with reason: %s",  iscsi_reject_reason_str(reason));
-		iscsi_logout_async_internal(iscsi, iscsi_reconnect_after_logout, NULL, ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_URGENT_DELIVERY);
+		iscsi_logout_async(iscsi, iscsi_reconnect_after_logout, NULL);
 		return 0;
 	}
 
@@ -379,9 +348,7 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 
 	itt = scsi_get_uint32(&in->data[16]);
 
-	if (iscsi->log_level > 1) {
-		iscsi_dump_pdu_header(iscsi, in->data);
-	}
+	iscsi_dump_pdu_header(iscsi, in->data);
 
 	for (pdu = iscsi->waitpdu; pdu; pdu = pdu->next) {
 		if (pdu->itt == itt) {
@@ -398,7 +365,7 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 
 	if (pdu->callback) {
 		pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-					pdu->private_data);
+		              pdu->private_data);
 	}
 
 	ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
@@ -407,44 +374,97 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 }
 
 
+static void iscsi_process_pdu_serials(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
+{
+	uint32_t itt = scsi_get_uint32(&in->hdr[16]);
+	uint32_t statsn = scsi_get_uint32(&in->hdr[24]);
+	uint32_t maxcmdsn = scsi_get_uint32(&in->hdr[32]);
+	uint32_t expcmdsn = scsi_get_uint32(&in->hdr[28]);
+	uint16_t status = scsi_get_uint16(&in->hdr[36]);
+	uint8_t flags = in->hdr[1];
+	enum iscsi_opcode opcode = in->hdr[0] & 0x3f;
+
+	/* RFC3720 10.13.5 (serials are invalid if status class != 0) */
+	if (opcode == ISCSI_PDU_LOGIN_RESPONSE && (status >> 8)) {
+		return;
+	}
+
+	if (iscsi_serial32_compare(maxcmdsn, iscsi->maxcmdsn) > 0) {
+		iscsi->maxcmdsn = maxcmdsn;
+	}
+	if (iscsi_serial32_compare(expcmdsn, iscsi->expcmdsn) > 0) {
+		iscsi->expcmdsn = expcmdsn;
+	}
+
+	/* RFC3720 10.7.3 (StatSN is invalid if S bit unset in flags) */
+	if (opcode == ISCSI_PDU_DATA_IN &&
+	    !(flags & ISCSI_PDU_DATA_CONTAINS_STATUS)) {
+		return;
+	}
+
+	if (itt == 0xffffffff) {
+		/* target will not increase statsn if itt == 0xffffffff */
+		statsn--;
+	}
+	if (iscsi_serial32_compare(statsn, iscsi->statsn) > 0) {
+		iscsi->statsn = statsn;
+	}
+}
+
 int
 iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 {
-	uint32_t itt;
-	enum iscsi_opcode opcode;
+	uint32_t itt = scsi_get_uint32(&in->hdr[16]);
+	enum iscsi_opcode opcode = in->hdr[0] & 0x3f;
+	uint8_t ahslen = in->hdr[4];
 	struct iscsi_pdu *pdu;
-	uint8_t	ahslen;
-
-	opcode = in->hdr[0] & 0x3f;
-	ahslen = in->hdr[4];
-	itt = scsi_get_uint32(&in->hdr[16]);
 
 	if (ahslen != 0) {
 		iscsi_set_error(iscsi, "cant handle expanded headers yet");
 		return -1;
 	}
 
+	/* All target PDUs update the serials */
+	iscsi_process_pdu_serials(iscsi, in);
+
 	if (opcode == ISCSI_PDU_ASYNC_MSG) {
 		uint8_t event = in->hdr[36];
 		uint16_t param1 = scsi_get_uint16(&in->hdr[38]); 
 		uint16_t param2 = scsi_get_uint16(&in->hdr[40]); 
 		uint16_t param3 = scsi_get_uint16(&in->hdr[42]); 
 		switch (event) {
+		case 0x0:
+			/* Just ignore these ones for now. It could be
+			 * a UNIT_ATTENTION for some changes on the
+			 * target but we don't have an API to pass this on
+			 * to the application yet.
+			 */
+			ISCSI_LOG(iscsi, 2, "Ignoring received iSCSI AsyncMsg/"
+				  "SCSI Async Event");
+			return 0;
 		case 0x1:
 			ISCSI_LOG(iscsi, 2, "target requests logout within %u seconds", param3);
-			iscsi_logout_async_internal(iscsi, iscsi_reconnect_after_logout, NULL, ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_URGENT_DELIVERY);
+			/* this is an ugly workaround for DELL Equallogic FW 7.x bugs:
+			 *  Bug_71409 - I/O errors during volume move (present before 7.0.7)
+			 *  Bug_73732 - I/O errors during volume move operation (still present in 7.0.9)
+			 */
+			if (getenv("LIBISCSI_DROP_CONN_ON_ASYNC_EVENT1") != NULL) {
+				ISCSI_LOG(iscsi, 2, "dropping connection to fix errors with broken DELL Equallogic firmware 7.x");
+				return -1;
+			}
+			iscsi_logout_async(iscsi, iscsi_reconnect_after_logout, NULL);
 			return 0;
 		case 0x2:
 			ISCSI_LOG(iscsi, 2, "target will drop this connection. Time2Wait is %u seconds", param2);
-			iscsi->last_reconnect = time(NULL) + param2;
+			iscsi->next_reconnect = time(NULL) + param2;
 			return 0;
 		case 0x3:
 			ISCSI_LOG(iscsi, 2, "target will drop all connections of this session. Time2Wait is %u seconds", param2);
-			iscsi->last_reconnect = time(NULL) + param2;
+			iscsi->next_reconnect = time(NULL) + param2;
 			return 0;
 		case 0x4:
 			ISCSI_LOG(iscsi, 2, "target requests parameter renogitiation.");
-			iscsi_logout_async_internal(iscsi, iscsi_reconnect_after_logout, NULL, ISCSI_PDU_DROP_ON_RECONNECT);
+			iscsi_logout_async(iscsi, iscsi_reconnect_after_logout, NULL);
 			return 0;
 		default:
 			ISCSI_LOG(iscsi, 1, "unhandled async event %u: param1 %u param2 %u param3 %u", event, param1, param2, param3);
@@ -618,6 +638,7 @@ void
 iscsi_pdu_set_cmdsn(struct iscsi_pdu *pdu, uint32_t cmdsn)
 {
 	scsi_set_uint32(&pdu->outdata.data[24], cmdsn);
+	pdu->cmdsn = cmdsn;
 }
 
 void
@@ -654,12 +675,61 @@ iscsi_pdu_set_cdb(struct iscsi_pdu *pdu, struct scsi_task *task)
 void
 iscsi_pdu_set_lun(struct iscsi_pdu *pdu, uint32_t lun)
 {
-	pdu->outdata.data[8] = lun >> 8;
-	pdu->outdata.data[9] = lun & 0xff;
+	scsi_set_uint16(&pdu->outdata.data[8], lun);
 }
 
 void
 iscsi_pdu_set_expxferlen(struct iscsi_pdu *pdu, uint32_t expxferlen)
 {
+	pdu->expxferlen = expxferlen;
 	scsi_set_uint32(&pdu->outdata.data[20], expxferlen);
 }
+
+void
+iscsi_timeout_scan(struct iscsi_context *iscsi)
+{
+	struct iscsi_pdu *pdu;
+	struct iscsi_pdu *next_pdu;
+	time_t t = time(NULL);
+
+	for (pdu = iscsi->outqueue; pdu; pdu = next_pdu) {
+		next_pdu = pdu->next;
+
+		if (pdu->scsi_timeout == 0) {
+			/* no timeout for this pdu */
+			continue;
+		}
+		if (t < pdu->scsi_timeout) {
+			/* not expired yet */
+			continue;
+		}
+		ISCSI_LIST_REMOVE(&iscsi->outqueue, pdu);
+		iscsi_set_error(iscsi, "command timed out");
+		iscsi_dump_pdu_header(iscsi, pdu->outdata.data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_TIMEOUT,
+			              NULL, pdu->private_data);
+		}
+		iscsi_free_pdu(iscsi, pdu);
+	}
+	for (pdu = iscsi->waitpdu; pdu; pdu = next_pdu) {
+		next_pdu = pdu->next;
+
+		if (pdu->scsi_timeout == 0) {
+			/* no timeout for this pdu */
+			continue;
+		}
+		if (t < pdu->scsi_timeout) {
+			/* not expired yet */
+			continue;
+		}
+		ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
+		iscsi_set_error(iscsi, "command timed out");
+		iscsi_dump_pdu_header(iscsi, pdu->outdata.data);
+		if (pdu->callback) {
+			pdu->callback(iscsi, SCSI_STATUS_TIMEOUT,
+			              NULL, pdu->private_data);
+		}
+		iscsi_free_pdu(iscsi, pdu);
+	}
+}
