@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
@@ -92,10 +93,17 @@ iscsi_add_to_outqueue(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 	
 	/* queue pdus in ascending order of CmdSN. 
 	 * ensure that pakets with the same CmdSN are kept in FIFO order.
+	 * immediate PDUs are queued in front of queue with the CmdSN
+	 * of the first element in the outqueue.
 	 */
+
+	if (pdu->outdata.data[0] & ISCSI_PDU_IMMEDIATE) {
+		iscsi_pdu_set_cmdsn(pdu, current->cmdsn);
+	}
+
 	do {
 		if (iscsi_serial32_compare(pdu->cmdsn, current->cmdsn) < 0 ||
-			pdu->flags & ISCSI_PDU_URGENT_DELIVERY) {
+			(pdu->outdata.data[0] & ISCSI_PDU_IMMEDIATE && !(current->outdata.data[0] & ISCSI_PDU_IMMEDIATE))) {
 			/* insert PDU before the current */
 			if (last != NULL) {
 				last->next=pdu;
@@ -129,11 +137,11 @@ static void set_nonblocking(int fd)
 #endif
 }
 
-int set_tcp_sockopt(int sockfd, int optname, int value)
+static int set_tcp_sockopt(int sockfd, int optname, int value)
 {
 	int level;
 
-	#if defined(__FreeBSD__) || defined(__sun) || (defined(__APPLE__) && defined(__MACH__))
+	#ifndef SOL_TCP
 	struct protoent *buf;
 
 	if ((buf = getprotobyname("tcp")) != NULL)
@@ -151,7 +159,7 @@ int set_tcp_sockopt(int sockfd, int optname, int value)
 #define TCP_USER_TIMEOUT        18
 #endif
 
-int set_tcp_user_timeout(struct iscsi_context *iscsi)
+static int set_tcp_user_timeout(struct iscsi_context *iscsi)
 {
 	if (set_tcp_sockopt(iscsi->fd, TCP_USER_TIMEOUT, iscsi->tcp_user_timeout) != 0) {
 		iscsi_set_error(iscsi, "TCP: Failed to set tcp user timeout. Error %s(%d)", strerror(errno), errno);
@@ -165,7 +173,7 @@ int set_tcp_user_timeout(struct iscsi_context *iscsi)
 #define TCP_SYNCNT        7
 #endif
 
-int set_tcp_syncnt(struct iscsi_context *iscsi)
+static int set_tcp_syncnt(struct iscsi_context *iscsi)
 {
 	if (set_tcp_sockopt(iscsi->fd, TCP_SYNCNT, iscsi->tcp_syncnt) != 0) {
 		iscsi_set_error(iscsi, "TCP: Failed to set tcp syn retries. Error %s(%d)", strerror(errno), errno);
@@ -284,6 +292,14 @@ iscsi_connect_async(struct iscsi_context *iscsi, const char *portal,
 
 	}
 
+	if (iscsi->old_iscsi && iscsi->fd != iscsi->old_iscsi->fd) {
+		if (dup2(iscsi->fd, iscsi->old_iscsi->fd) == -1) {
+			return -1;
+		}
+		close(iscsi->fd);
+		iscsi->fd = iscsi->old_iscsi->fd;
+	}
+
 	iscsi->socket_status_cb  = cb;
 	iscsi->connect_data      = private_data;
 
@@ -362,12 +378,14 @@ iscsi_disconnect(struct iscsi_context *iscsi)
 
 	close(iscsi->fd);
 
-	if (iscsi->connected_portal[0]) {
+	if (!(iscsi->pending_reconnect && iscsi->old_iscsi) &&
+	    iscsi->connected_portal[0]) {
 		ISCSI_LOG(iscsi, 2, "disconnected from portal %s",iscsi->connected_portal);
 	}
 
 	iscsi->fd  = -1;
 	iscsi->is_connected = 0;
+	iscsi->is_corked = 0;
 
 	return 0;
 }
@@ -375,6 +393,9 @@ iscsi_disconnect(struct iscsi_context *iscsi)
 int
 iscsi_get_fd(struct iscsi_context *iscsi)
 {
+	if (iscsi->old_iscsi) {
+		return iscsi->old_iscsi->fd;
+	}
 	return iscsi->fd;
 }
 
@@ -383,9 +404,17 @@ iscsi_which_events(struct iscsi_context *iscsi)
 {
 	int events = iscsi->is_connected ? POLLIN : POLLOUT;
 
+	if (iscsi->pending_reconnect && iscsi->old_iscsi &&
+		time(NULL) < iscsi->next_reconnect) {
+		return 0;
+	}
+
 	if (iscsi->outqueue_current != NULL ||
 	    (iscsi->outqueue != NULL && !iscsi->is_corked &&
-	     iscsi_serial32_compare(iscsi->outqueue->cmdsn, iscsi->maxcmdsn) <= 0)) {
+	     (iscsi_serial32_compare(iscsi->outqueue->cmdsn, iscsi->maxcmdsn) <= 0 ||
+	      iscsi->outqueue->outdata.data[0] & ISCSI_PDU_IMMEDIATE)
+	    )
+	   ) {
 		events |= POLLOUT;
 	}
 	return events;
@@ -410,18 +439,38 @@ iscsi_queue_length(struct iscsi_context *iscsi)
 	return i;
 }
 
+int
+iscsi_out_queue_length(struct iscsi_context *iscsi)
+{
+	int i = 0;
+	struct iscsi_pdu *pdu;
+
+	for (pdu = iscsi->outqueue; pdu; pdu = pdu->next) {
+		i++;
+	}
+
+	return i;
+}
+
 ssize_t
 iscsi_iovector_readv_writev(struct iscsi_context *iscsi, struct scsi_iovector *iovector, uint32_t pos, ssize_t count, int do_write)
 {
-	if (iovector->iov == NULL) {
+        struct scsi_iovec *iov, *iov2;
+        int niov;
+        uint32_t len2;
+        size_t _len2;
+        ssize_t n;
+
+        if (iovector->iov == NULL) {
 		errno = EINVAL;
 		return -1;
 	}
 
 	if (pos < iovector->offset) {
-		/* start over in case we are going backwards */
-		iovector->offset = 0;
-		iovector->consumed = 0;
+		iscsi_set_error(iscsi, "iovector reset. pos is smaller than"
+				"current offset");
+		errno = EINVAL;
+		return -1;
 	}
 
 	if (iovector->niov <= iovector->consumed) {
@@ -433,7 +482,7 @@ iscsi_iovector_readv_writev(struct iscsi_context *iscsi, struct scsi_iovector *i
 	}
 
 	/* iov is a pointer to the first iovec to pass */
-	struct scsi_iovec *iov = &iovector->iov[iovector->consumed];
+	iov = &iovector->iov[iovector->consumed];
 	pos -= iovector->offset;
 
 	/* forward until iov points to the first iov to pass */
@@ -448,34 +497,31 @@ iscsi_iovector_readv_writev(struct iscsi_context *iscsi, struct scsi_iovector *i
 		iov = &iovector->iov[iovector->consumed];
 	}
 
-	/* iov2 is a pointer to the last iovec to pass */
-	struct scsi_iovec *iov2 = iov;
+	iov2 = iov;         /* iov2 is a pointer to the last iovec to pass */
+	niov = 1;           /* number of iovectors to pass */
+	len2 = pos + count; /* adjust length of iov2 */
 
-	int niov=1; /* number of iovectors to pass */
-	uint32_t len2 = pos + count; /* adjust length of iov2 */
-	
 	/* forward until iov2 points to the last iovec we pass later. it might
 	   happen that we have a lot of iovectors but are limited by count */
 	while (len2 > iov2->iov_len) {
-		if (iovector->niov <= iovector->consumed+niov-1) {
+		niov++;
+		if (iovector->niov < iovector->consumed + niov) {
 			errno = EINVAL;
 			return -1;
 		}
-		niov++;
 		len2 -= iov2->iov_len;
-		iov2 = &iovector->iov[iovector->consumed+niov-1];
+		iov2 = &iovector->iov[iovector->consumed + niov - 1];
 	}
 
 	/* we might limit the length of the last iovec we pass to readv/writev
 	   store its orignal length to restore it later */
-	size_t _len2 = iov2->iov_len;
+	_len2 = iov2->iov_len;
 
 	/* adjust base+len of start iovec and len of last iovec */
 	iov2->iov_len = len2;
 	iov->iov_base = (void*) ((uintptr_t)iov->iov_base + pos);
 	iov->iov_len -= pos;
 
-	ssize_t n;
 	if (do_write) {
 		n = writev(iscsi->fd, (struct iovec*) iov, niov);
 	} else {
@@ -577,7 +623,8 @@ iscsi_read_from_socket(struct iscsi_context *iscsi)
 				return 0;
 			}
 			iscsi_set_error(iscsi, "read from socket failed, "
-				"errno:%d", errno);
+					"errno:%d %s", errno,
+					iscsi_get_error(iscsi));
 			return -1;
 		}
 
@@ -613,6 +660,13 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 	size_t total;
 	struct iscsi_pdu *pdu;
 	static char padding_buf[3];
+	int socket_flags = 0;
+
+#ifdef MSG_NOSIGNAL
+	socket_flags |= MSG_NOSIGNAL;
+#elif SO_NOSIGPIPE
+	socket_flags |= SO_NOSIGPIPE;
+#endif
 
 	if (iscsi->fd == -1) {
 		iscsi_set_error(iscsi, "trying to write but not connected");
@@ -624,16 +678,31 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 			if (iscsi->is_corked) {
 				/* connection is corked we are not allowed to send
 				 * additional PDUs */
+				ISCSI_LOG(iscsi, 6, "iscsi_write_to_socket: socket is corked");
 				return 0;
 			}
 			
 			if (iscsi_serial32_compare(iscsi->outqueue->cmdsn, iscsi->maxcmdsn) > 0
 				&& !(iscsi->outqueue->outdata.data[0] & ISCSI_PDU_IMMEDIATE)) {
 				/* stop sending for non-immediate PDUs. maxcmdsn is reached */
+				ISCSI_LOG(iscsi, 6,
+				          "iscsi_write_to_socket: maxcmdsn reached (outqueue[0]->cmdsnd %08x > maxcmdsn %08x)",
+				          iscsi->outqueue->cmdsn, iscsi->maxcmdsn);
 				return 0;
 			}
+
 			/* pop first element of the outqueue */
+			if (iscsi_serial32_compare(iscsi->outqueue->cmdsn, iscsi->expcmdsn) < 0 &&
+				(iscsi->outqueue->outdata.data[0] & 0x3f) != ISCSI_PDU_DATA_OUT) {
+				iscsi_set_error(iscsi, "iscsi_write_to_socket: outqueue[0]->cmdsn < expcmdsn (%08x < %08x) opcode %02x",
+				                iscsi->outqueue->cmdsn, iscsi->expcmdsn, iscsi->outqueue->outdata.data[0] & 0x3f);
+				return -1;
+			}
 			iscsi->outqueue_current = iscsi->outqueue;
+			
+			/* set exp statsn */
+			iscsi_pdu_set_expstatsn(iscsi->outqueue_current, iscsi->statsn + 1);
+			
 			ISCSI_LIST_REMOVE(&iscsi->outqueue, iscsi->outqueue_current);
 			if (!(iscsi->outqueue_current->flags & ISCSI_PDU_DELETE_WHEN_SENT)) {
 				/* we have to add the pdu to the waitqueue already here
@@ -652,7 +721,7 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 			count = send(iscsi->fd,
 				     pdu->outdata.data + pdu->outdata_written,
 				     pdu->outdata.size - pdu->outdata_written,
-				     0);
+				     socket_flags);
 			if (count == -1) {
 				if (errno == EAGAIN || errno == EWOULDBLOCK) {
 					return 0;
@@ -688,7 +757,8 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 					return 0;
 				}
 				iscsi_set_error(iscsi, "Error when writing to "
-						"socket :%d", errno);
+						"socket :%d %s", errno,
+						iscsi_get_error(iscsi));
 				return -1;
 			}
 
@@ -700,7 +770,7 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 
 		/* Write padding */
 		if (pdu->payload_written < total) {
-			count = send(iscsi->fd, padding_buf, total - pdu->payload_written, 0);
+			count = send(iscsi->fd, padding_buf, total - pdu->payload_written, socket_flags);
 			if (count == -1) {
 				if (errno == EAGAIN || errno == EWOULDBLOCK) {
 					return 0;
@@ -715,12 +785,12 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 		if (pdu->payload_written != total) {
 			return 0;
 		}
-		if (pdu->flags & ISCSI_PDU_DELETE_WHEN_SENT) {
-			iscsi_free_pdu(iscsi, pdu);
-		}
 		if (pdu->flags & ISCSI_PDU_CORK_WHEN_SENT) {
 			iscsi->is_corked = 1;
 		}
+		if (pdu->flags & ISCSI_PDU_DELETE_WHEN_SENT) {
+			iscsi_free_pdu(iscsi, pdu);
+		}
 		iscsi->outqueue_current = NULL;
 	}
 	return 0;
@@ -734,6 +804,14 @@ iscsi_service_reconnect_if_loggedin(struct iscsi_context *iscsi)
 			return 0;
 		}
 	}
+	if (iscsi->old_iscsi) {
+		if (!iscsi->pending_reconnect) {
+			iscsi_reconnect_cb(iscsi, SCSI_STATUS_ERROR, NULL, NULL);
+		}
+		return 0;
+	}
+	iscsi_set_error(iscsi, "iscsi_service_reconnect_if_loggedin. Can not "
+			"reconnect right now.\n");
 	return -1;
 }
 
@@ -745,7 +823,13 @@ iscsi_service(struct iscsi_context *iscsi, int revents)
 	}
 
 	if (iscsi->pending_reconnect) {
-		iscsi_reconnect(iscsi);
+		if (time(NULL) >= iscsi->next_reconnect) {
+			return iscsi_reconnect(iscsi);
+		} else {
+			if (iscsi->old_iscsi) {
+				return 0;
+			}
+		}
 	}
 
 	if (revents & POLLERR) {
@@ -819,16 +903,17 @@ iscsi_service(struct iscsi_context *iscsi, int revents)
 		return 0;
 	}
 
-	if (revents & POLLOUT && (iscsi->outqueue != NULL || iscsi->outqueue_current != NULL)) {
-		if (iscsi_write_to_socket(iscsi) != 0) {
-			return iscsi_service_reconnect_if_loggedin(iscsi);
-		}
-	}
 	if (revents & POLLIN) {
 		if (iscsi_read_from_socket(iscsi) != 0) {
 			return iscsi_service_reconnect_if_loggedin(iscsi);
 		}
 	}
+	if (revents & POLLOUT) {
+		if (iscsi_write_to_socket(iscsi) != 0) {
+			return iscsi_service_reconnect_if_loggedin(iscsi);
+		}
+	}
+	iscsi_timeout_scan(iscsi);
 
 	return 0;
 }
