@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
@@ -35,6 +36,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <unistd.h>
 #include "iscsi.h"
 #include "iscsi-private.h"
 #include "scsi-lowlevel.h"
@@ -52,6 +54,8 @@ event_loop(struct iscsi_context *iscsi, struct iscsi_sync_state *state)
 	int ret;
 
 	while (state->finished == 0) {
+		short revents;
+
 		pfd.fd = iscsi_get_fd(iscsi);
 		pfd.events = iscsi_which_events(iscsi);
 
@@ -60,11 +64,8 @@ event_loop(struct iscsi_context *iscsi, struct iscsi_sync_state *state)
 			state->status = -1;
 			return;
 		}
-		if (ret == 0) {
-			iscsi_timeout_scan(iscsi);
-			continue;
-		}
-		if (iscsi_service(iscsi, pfd.revents) < 0) {
+		revents = (ret == 0) ? 0 : pfd.revents;
+		if (iscsi_service(iscsi, revents) < 0) {
 			iscsi_set_error(iscsi,
 				"iscsi_service failed with : %s",
 				iscsi_get_error(iscsi));
@@ -167,16 +168,61 @@ int iscsi_logout_sync(struct iscsi_context *iscsi)
 	return state.status;
 }
 
+static void
+reconnect_event_loop(struct iscsi_context *iscsi, struct iscsi_sync_state *state)
+{
+	struct pollfd pfd;
+	int ret;
+	while (iscsi->old_iscsi) {
+		pfd.fd = iscsi_get_fd(iscsi);
+		pfd.events = iscsi_which_events(iscsi);
+
+		if (!pfd.events) {
+			sleep(1);
+			continue;
+		}
+
+		if ((ret = poll(&pfd, 1, 1000)) < 0) {
+			iscsi_set_error(iscsi, "Poll failed");
+			state->status = -1;
+			return;
+		}
+
+		if (iscsi_service(iscsi, pfd.revents) < 0) {
+			iscsi_set_error(iscsi,
+				"iscsi_service failed with : %s",
+				iscsi_get_error(iscsi));
+			state->status = -1;
+			return;
+		}
+	}
+	state->status = 0;
+}
+
+int iscsi_reconnect_sync(struct iscsi_context *iscsi)
+{
+	struct iscsi_sync_state state;
+
+	memset(&state, 0, sizeof(state));
+
+	if (iscsi_reconnect(iscsi) != 0) {
+		iscsi_set_error(iscsi, "Failed to reconnect. %s", iscsi_get_error(iscsi));
+		return -1;
+	}
+
+	reconnect_event_loop(iscsi, &state);
+
+	return state.status;
+}
+
 static void
 iscsi_task_mgmt_sync_cb(struct iscsi_context *iscsi, int status,
 	      void *command_data, void *private_data)
 {
 	struct iscsi_sync_state *state = private_data;
 
-	if (state != NULL) {
-		state->status    = status;
-		state->finished = 1;
-	}
+	state->status   = status;
+	state->finished = 1;
 
 	/* The task mgmt command might have completed successfully
 	 * but the target might have responded with
@@ -200,7 +246,7 @@ iscsi_task_mgmt_sync_cb(struct iscsi_context *iscsi, int status,
 			break;
 		}
 
-		state->status = SCSI_STATUS_ERROR;
+                state->status = SCSI_STATUS_ERROR;
 	}
 }
 
@@ -230,8 +276,6 @@ int
 iscsi_task_mgmt_abort_task_sync(struct iscsi_context *iscsi,
 				struct scsi_task *task)
 {
-	iscsi_scsi_cancel_task(iscsi, task);
-
 	return iscsi_task_mgmt_sync(iscsi, task->lun,
 				    ISCSI_TM_ABORT_TASK,
 				    task->itt, task->cmdsn);
@@ -779,6 +823,29 @@ iscsi_write16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 	return state.task;
 }
 
+struct scsi_task *
+iscsi_writeatomic16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			 unsigned char *data, uint32_t datalen, int blocksize,
+			 int wrprotect, int dpo, int fua, int group_number)
+{
+	struct iscsi_sync_state state;
+
+	memset(&state, 0, sizeof(state));
+
+	if (iscsi_writeatomic16_task(iscsi, lun, lba,
+				     data, datalen, blocksize, wrprotect,
+				     dpo, fua, group_number,
+				     scsi_sync_cb, &state) == NULL) {
+		iscsi_set_error(iscsi,
+				"Failed to send WriteAtomic16 command");
+		return NULL;
+	}
+
+	event_loop(iscsi, &state);
+
+	return state.task;
+}
+
 struct scsi_task *
 iscsi_orwrite_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
