@@ -17,8 +17,8 @@
    along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
 
-#ifndef	_ISCSI_TEST_CU_H_
-#define	_ISCSI_TEST_CU_H_
+#ifndef        _ISCSI_TEST_CU_H_
+#define        _ISCSI_TEST_CU_H_
 
 #include <time.h>
 #include <sys/types.h>
@@ -32,11 +32,7 @@
 #include "iscsi-support.h"
 
 /* globals between setup, tests, and teardown */
-extern struct iscsi_context *iscsic;
-extern int tgt_lun;
 extern struct scsi_task *task;
-extern struct iscsi_context *iscsic2;
-extern int tgt_lun2;
 extern unsigned char *read_write_buf;
 
 #ifndef HAVE_CU_SUITEINFO_PSETUPFUNC
@@ -53,8 +49,16 @@ void test_setup(void);
 void test_teardown(void);
 
 void test_compareandwrite_simple(void);
+void test_compareandwrite_dpofua(void);
 void test_compareandwrite_miscompare(void);
 
+void test_extendedcopy_simple(void);
+void test_extendedcopy_param(void);
+void test_extendedcopy_descr_limits(void);
+void test_extendedcopy_descr_type(void);
+void test_extendedcopy_validate_tgt_descr(void);
+void test_extendedcopy_validate_seg_descr(void);
+
 void test_get_lba_status_simple(void);
 void test_get_lba_status_beyond_eol(void);
 void test_get_lba_status_unmap_single(void);
@@ -75,6 +79,9 @@ void test_iscsi_datasn_invalid(void);
 void test_mandatory_sbc(void);
 
 void test_modesense6_all_pages(void);
+void test_modesense6_control(void);
+void test_modesense6_control_d_sense(void);
+void test_modesense6_control_swp(void);
 void test_modesense6_residuals(void);
 
 void test_nomedia_sbc(void);
@@ -83,7 +90,7 @@ void test_orwrite_simple(void);
 void test_orwrite_beyond_eol(void);
 void test_orwrite_0blocks(void);
 void test_orwrite_wrprotect(void);
-void test_orwrite_flags(void);
+void test_orwrite_dpofua(void);
 void test_orwrite_verify(void);
 
 void test_prefetch10_simple(void);
@@ -107,6 +114,7 @@ void test_preventallow_2_itnexuses(void);
 
 void test_prin_read_keys_simple(void);
 void test_prin_serviceaction_range(void);
+void test_prin_report_caps_simple(void);
 
 void test_prout_register_simple(void);
 void test_prout_reserve_simple(void);
@@ -122,6 +130,8 @@ void test_prout_reserve_ownership_earo(void);
 void test_prout_reserve_ownership_wero(void);
 void test_prout_reserve_ownership_eaar(void);
 void test_prout_reserve_ownership_wear(void);
+void test_prout_clear_simple(void);
+void test_prout_preempt_rm_reg(void);
 
 void test_read6_simple(void);
 void test_read6_beyond_eol(void);
@@ -132,22 +142,23 @@ void test_read10_simple(void);
 void test_read10_beyond_eol(void);
 void test_read10_0blocks(void);
 void test_read10_rdprotect(void);
-void test_read10_flags(void);
+void test_read10_dpofua(void);
 void test_read10_residuals(void);
 void test_read10_invalid(void);
+void test_async_read(void);
 
 void test_read12_simple(void);
 void test_read12_beyond_eol(void);
 void test_read12_0blocks(void);
 void test_read12_rdprotect(void);
-void test_read12_flags(void);
+void test_read12_dpofua(void);
 void test_read12_residuals(void);
 
 void test_read16_simple(void);
 void test_read16_beyond_eol(void);
 void test_read16_0blocks(void);
 void test_read16_rdprotect(void);
-void test_read16_flags(void);
+void test_read16_dpofua(void);
 void test_read16_residuals(void);
 
 void test_readcapacity10_simple(void);
@@ -155,9 +166,13 @@ void test_readcapacity10_simple(void);
 void test_readcapacity16_alloclen(void);
 void test_readcapacity16_protection(void);
 void test_readcapacity16_simple(void);
+void test_readcapacity16_support(void);
 
 void test_readonly_sbc(void);
 
+void test_receive_copy_results_copy_status(void);
+void test_receive_copy_results_op_params(void);
+
 void test_report_supported_opcodes_one_command(void);
 void test_report_supported_opcodes_rctd(void);
 void test_report_supported_opcodes_servactv(void);
@@ -198,6 +213,7 @@ void test_verify10_beyond_eol(void);
 void test_verify10_0blocks(void);
 void test_verify10_vrprotect(void);
 void test_verify10_flags(void);
+void test_verify10_dpo(void);
 void test_verify10_mismatch(void);
 void test_verify10_mismatch_no_cmp(void);
 
@@ -206,6 +222,7 @@ void test_verify12_beyond_eol(void);
 void test_verify12_0blocks(void);
 void test_verify12_vrprotect(void);
 void test_verify12_flags(void);
+void test_verify12_dpo(void);
 void test_verify12_mismatch(void);
 void test_verify12_mismatch_no_cmp(void);
 
@@ -214,6 +231,7 @@ void test_verify16_beyond_eol(void);
 void test_verify16_0blocks(void);
 void test_verify16_vrprotect(void);
 void test_verify16_flags(void);
+void test_verify16_dpo(void);
 void test_verify16_mismatch(void);
 void test_verify16_mismatch_no_cmp(void);
 
@@ -221,23 +239,31 @@ void test_write10_simple(void);
 void test_write10_beyond_eol(void);
 void test_write10_0blocks(void);
 void test_write10_wrprotect(void);
-void test_write10_flags(void);
+void test_write10_dpofua(void);
 void test_write10_residuals(void);
+void test_async_write(void);
 
 void test_write12_simple(void);
 void test_write12_beyond_eol(void);
 void test_write12_0blocks(void);
 void test_write12_wrprotect(void);
-void test_write12_flags(void);
+void test_write12_dpofua(void);
 void test_write12_residuals(void);
 
 void test_write16_simple(void);
 void test_write16_beyond_eol(void);
 void test_write16_0blocks(void);
 void test_write16_wrprotect(void);
-void test_write16_flags(void);
+void test_write16_dpofua(void);
 void test_write16_residuals(void);
 
+void test_writeatomic16_simple(void);
+void test_writeatomic16_beyond_eol(void);
+void test_writeatomic16_0blocks(void);
+void test_writeatomic16_wrprotect(void);
+void test_writeatomic16_dpofua(void);
+void test_writeatomic16_vpd(void);
+
 void test_writesame10_simple(void);
 void test_writesame10_beyond_eol(void);
 void test_writesame10_0blocks(void);
@@ -261,6 +287,7 @@ void test_writeverify10_beyond_eol(void);
 void test_writeverify10_0blocks(void);
 void test_writeverify10_wrprotect(void);
 void test_writeverify10_flags(void);
+void test_writeverify10_dpo(void);
 void test_writeverify10_residuals(void);
 
 void test_writeverify12_simple(void);
@@ -268,6 +295,7 @@ void test_writeverify12_beyond_eol(void);
 void test_writeverify12_0blocks(void);
 void test_writeverify12_wrprotect(void);
 void test_writeverify12_flags(void);
+void test_writeverify12_dpo(void);
 void test_writeverify12_residuals(void);
 
 void test_writeverify16_simple(void);
@@ -275,6 +303,14 @@ void test_writeverify16_beyond_eol(void);
 void test_writeverify16_0blocks(void);
 void test_writeverify16_wrprotect(void);
 void test_writeverify16_flags(void);
+void test_writeverify16_dpo(void);
 void test_writeverify16_residuals(void);
 
-#endif	/* _ISCSI_TEST_CU_H_ */
+void test_multipathio_simple(void);
+void test_multipathio_reset(void);
+void test_multipathio_compareandwrite(void);
+void test_mpio_async_caw(void);
+
+void test_async_abort_simple(void);
+
+#endif        /* _ISCSI_TEST_CU_H_ */
