@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /* 
    Copyright (C) 2013 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
    
@@ -28,17 +29,9 @@
 void
 test_readcapacity16_simple(void)
 {
-	int ret;
+        logging(LOG_VERBOSE, LOG_BLANK_LINE);
+        logging(LOG_VERBOSE, "Test that READCAPACITY16 works");
 
-
-	logging(LOG_VERBOSE, LOG_BLANK_LINE);
-	logging(LOG_VERBOSE, "Test that READCAPACITY16 works");
-
-	ret = readcapacity16(iscsic, tgt_lun, 16);
-	if (ret == -2) {
-		logging(LOG_NORMAL, "[SKIPPED] READCAPACITY16 is not implemented on this target and it does not claim support.");
-		CU_PASS("READCAPACITY16 is not implemented and no SBC-3 support claimed.");
-		return;
-	}	
-	CU_ASSERT_EQUAL(ret, 0);
+        READCAPACITY16(sd, NULL, 16,
+                       EXPECT_STATUS_GOOD);
 }
