@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /* 
    Copyright (C) 2013 Ronnie Sahlberg <ronniesahlberg@gmail.com>
    
@@ -28,32 +29,13 @@
 void
 test_verify10_flags(void)
 {
-	int ret;
-	unsigned char *buf = malloc(block_size);
+        logging(LOG_VERBOSE, LOG_BLANK_LINE);
+        logging(LOG_VERBOSE, "Test VERIFY10 flags");
 
-	logging(LOG_VERBOSE, LOG_BLANK_LINE);
-	logging(LOG_VERBOSE, "Test VERIFY10 flags");
-
-	ret = read10(iscsic, tgt_lun, 0, block_size,
-		     block_size, 0, 0, 0, 0, 0, buf);
-	CU_ASSERT_EQUAL(ret, 0);
-
-
-	logging(LOG_VERBOSE, "Test VERIFY10 with DPO==1");
-	ret = verify10(iscsic, tgt_lun, 0, block_size,
-		       block_size, 0, 1, 0, buf);
-	if (ret == -2) {
-		logging(LOG_NORMAL, "[SKIPPED] VERIFY10 is not implemented.");
-		CU_PASS("[SKIPPED] Target does not support VERIFY10. Skipping test");
-		free(buf);
-		return;
-	}
-	CU_ASSERT_EQUAL(ret, 0);
-
-
-	logging(LOG_VERBOSE, "Test VERIFY10 with BYTCHK==1");
-	ret = verify10(iscsic, tgt_lun, 0, block_size,
-		       block_size, 0, 0, 1, buf);
-	CU_ASSERT_EQUAL(ret, 0);
-	free(buf);
+        READ10(sd, NULL, 0, block_size,
+                     block_size, 0, 0, 0, 0, 0, scratch,
+                     EXPECT_STATUS_GOOD);
+        logging(LOG_VERBOSE, "Test VERIFY10 with BYTCHK==1");
+        VERIFY10(sd, 0, block_size, block_size, 0, 0, 1, scratch,
+                 EXPECT_STATUS_GOOD);
 }
