@@ -1,3 +1,4 @@
+/* -*-  mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:nil;  -*- */
 /* 
    Copyright (C) 2013 Ronnie Sahlberg <ronniesahlberg@gmail.com>
    
@@ -29,28 +30,14 @@
 void
 test_writeverify16_flags(void)
 { 
-	int ret;
-	unsigned char *buf = alloca(block_size);
+        CHECK_FOR_DATALOSS;
 
-	CHECK_FOR_DATALOSS;
+        logging(LOG_VERBOSE, LOG_BLANK_LINE);
+        logging(LOG_VERBOSE, "Test WRITEVERIFY16 flags");
 
-	logging(LOG_VERBOSE, LOG_BLANK_LINE);
-	logging(LOG_VERBOSE, "Test WRITEVERIFY16 flags");
 
-	logging(LOG_VERBOSE, "Test WRITEVERIFY16 with DPO==1");
-	ret = writeverify16(iscsic, tgt_lun, 0,
-		     block_size, block_size,
-		     0, 1, 0, 0, buf);
-	if (ret == -2) {
-		logging(LOG_NORMAL, "[SKIPPED] WRITE1VERIFY16 is not implemented.");
-		CU_PASS("WRITEVERIFY16 is not implemented.");
-		return;
-       	}	
-	CU_ASSERT_EQUAL(ret, 0);
-
-	logging(LOG_VERBOSE, "Test WRITEVERIFY16 with BYTCHK==1");
-	ret = writeverify16(iscsic, tgt_lun, 0,
-		     block_size, block_size,
-		     0, 0, 1, 0, buf);
-	CU_ASSERT_EQUAL(ret, 0);
+        logging(LOG_VERBOSE, "Test WRITEVERIFY16 with BYTCHK==1");
+        memset(scratch, 0xa6, block_size);
+        WRITEVERIFY16(sd, 0, block_size, block_size, 0, 0, 1, 0, scratch,
+                      EXPECT_STATUS_GOOD);
 }
